import logging

from spaceone.core import config
from spaceone.core.manager import BaseManager
from spaceone.core.connector.space_connector import SpaceConnector
from spaceone.core.auth.jwt.jwt_util import JWTUtil

_LOGGER = logging.getLogger(__name__)


class SecretManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        token = self.transaction.get_meta("token")
        self.token_type = JWTUtil.get_value_from_token(token, "typ")

        self.secret_conn: SpaceConnector = self.locator.get_connector(
            "SpaceConnector", service="secret"
        )

    def get_secret_data(self, secret_id: str, domain_id: str) -> dict:
        system_token = config.get_global("TOKEN")

        response = self.secret_conn.dispatch(
            "Secret.get_data",
            {"secret_id": secret_id, "domain_id": domain_id},
            token=system_token,
        )
        return response["data"]

    def get_trusted_secret_data(self, trusted_secret_id: str, domain_id: str) -> dict:
        system_token = config.get_global("TOKEN")

        response = self.secret_conn.dispatch(
            "TrustedSecret.get_data",
            {"trusted_secret_id": trusted_secret_id, "domain_id": domain_id},
            token=system_token,
        )
        return response["data"]

    def create_trusted_secret(self, params: dict) -> dict:
        return self.secret_conn.dispatch("TrustedSecret.create", params)

    def update_trusted_secret_data(
        self, trusted_secret_id: str, schema_id: str, data: dict
    ) -> None:
        self.secret_conn.dispatch(
            "TrustedSecret.update_data",
            {
                "trusted_secret_id": trusted_secret_id,
                "schema_id": schema_id,
                "data": data,
            },
        )

    def delete_related_trusted_secrets(self, trusted_account_id: str):
        response = self.list_trusted_secrets({"trusted_account_id": trusted_account_id})

        for trusted_secret_info in response.get("results", []):
            trusted_secret_id = trusted_secret_info["trusted_secret_id"]
            self.secret_conn.dispatch(
                "TrustedSecret.delete",
                {"trusted_secret_id": trusted_secret_id},
            )

    def list_trusted_secrets(self, params: dict) -> dict:
        return self.secret_conn.dispatch("TrustedSecret.list", params)

    def create_secret(self, params: dict, domain_id: str = None) -> dict:
        if self.token_type == "SYSTEM_TOKEN":
            return self.secret_conn.dispatch(
                "Secret.create", params, x_domain_id=domain_id
            )
        else:
            return self.secret_conn.dispatch("Secret.create", params)

    def update_secret(self, params: dict) -> dict:
        return self.secret_conn.dispatch("Secret.update", params)

    def delete_secret(self, secret_id: str, domain_id: str = None) -> None:
        if self.token_type == "SYSTEM_TOKEN":
            self.secret_conn.dispatch(
                "Secret.delete", {"secret_id": secret_id}, x_domain_id=domain_id
            )
        else:
            self.secret_conn.dispatch("Secret.delete", {"secret_id": secret_id})

    def update_secret_data(
        self,
        params: dict,
        domain_id: str = None,
        workspace_id: str = None,
    ) -> None:
        if self.token_type == "SYSTEM_TOKEN":
            self.secret_conn.dispatch(
                "Secret.update_data",
                params,
                x_domain_id=domain_id,
                x_workspace_id=workspace_id,
            )
        else:
            self.secret_conn.dispatch("Secret.update_data", params)

    def delete_related_secrets(self, service_account_id: str):
        response = self.list_secrets({"service_account_id": service_account_id})

        for secret_info in response.get("results", []):
            secret_id = secret_info["secret_id"]
            self.secret_conn.dispatch(
                "Secret.delete",
                {"secret_id": secret_id},
            )

    def list_secrets(self, params: dict, domain_id: str = None) -> dict:
        if self.token_type == "SYSTEM_TOKEN":
            return self.secret_conn.dispatch(
                "Secret.list", params, x_domain_id=domain_id
            )
        else:
            return self.secret_conn.dispatch("Secret.list", params)

    def create_user_secret(self, params: dict) -> dict:
        return self.secret_conn.dispatch("UserSecret.create", params)

    def get_user_secret_data(self, user_secret_id: str, domain_id: str = None) -> dict:
        response = self.secret_conn.dispatch("UserSecret.get_data", {"user_secret_id": user_secret_id, "domain_id": domain_id})
        return response["data"]

    def delete_user_secret(self, user_secret_id: str) -> None:
        self.secret_conn.dispatch("UserSecret.delete", {"user_secret_id": user_secret_id})

    def get_user_otp_secret_key(self, user_secret_id: str, domain_id: str = None) -> str:
        user_secret_info = self.get_user_secret_data(user_secret_id, domain_id)
        return user_secret_info["otp_secret_key"]



import logging
import pyotp
from collections import OrderedDict

from spaceone.core import utils, cache

from spaceone.identity.manager import SecretManager
from spaceone.identity.manager.mfa_manager.base import MFAManager

from spaceone.identity.error.error_user import ERROR_INVALID_VERIFY_CODE

_LOGGER = logging.getLogger(__name__)


class OTPMFAManager(MFAManager):
    mfa_type = "OTP"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def enable_mfa(self, user_id: str, domain_id: str, user_mfa: dict, user_vo):
        credentials = {
            "user_id": user_id,
            "domain_id": domain_id
        }

        otp_secret_key = self._generate_otp_secret_key()
        otp_qrcode_uri = self._generate_otp_qrcode_uri(self._generate_otp(otp_secret_key), user_id)

        self.set_cache_otp_mfa_secret_key(otp_secret_key, user_id, domain_id, credentials, user_mfa)

        user_mfa["options"]["otp_qrcode_uri"] = otp_qrcode_uri

        return user_mfa

    def disable_mfa(self, user_id: str, domain_id: str, user_mfa: dict, user_vo):
        credentials = {
            "user_id": user_id,
            "domain_id": domain_id
        }

        secret_manager: SecretManager = self.locator.get_manager(SecretManager)
        user_secret_id = user_mfa["options"].get("user_secret_id")
        otp_secret_key = secret_manager.get_user_otp_secret_key(user_secret_id, domain_id)

        self.set_cache_otp_mfa_secret_key(otp_secret_key, user_id, domain_id, credentials)

    def confirm_mfa(self, credentials: dict, verify_code: str):

        confirm_result = self.check_otp_mfa_verify_code(credentials, verify_code)

        return confirm_result

    def set_mfa_options(self, user_mfa: dict, credentials: dict):
        mfa_state = user_mfa.get("state", "DISABLED")

        secret_manager: SecretManager = self.locator.get_manager(SecretManager)

        if mfa_state == "ENABLED":
            user_secret_id = user_mfa["options"]["user_secret_id"]
            secret_manager.delete_user_secret(user_secret_id)

        elif mfa_state == "DISABLED":
            otp_secret_key = self.get_cached_otp_secret_key(credentials)
            user_secret_info = secret_manager.create_user_secret(otp_secret_key)
            user_mfa["options"]["user_secret_id"] = user_secret_info.get("user_secret_id")

        return user_mfa

    def set_cache_otp_mfa_secret_key(self, otp_secret_key: str, user_id: str, domain_id: str, credentials: dict, user_mfa: dict = None):
        if cache.is_set():
            ordered_credentials = OrderedDict(sorted(credentials.items()))
            hashed_credentials = utils.dict_to_hash(ordered_credentials)
            cache.delete(f"identity:mfa:{hashed_credentials}")
            cache.set(
                f"identity:mfa:{hashed_credentials}",
                {
                    "otp_secret_key": otp_secret_key,
                    "user_id": user_id,
                    "domain_id": domain_id,
                    "user_mfa": user_mfa
                },
                expire=self.CONST_MFA_VERIFICATION_CODE_TIMEOUT,
            )

    def check_otp_mfa_verify_code(self, credentials: dict, verify_code: str) -> bool:
        if cache.is_set():
            ordered_credentials = OrderedDict(sorted(credentials.items()))
            hashed_credentials = utils.dict_to_hash(ordered_credentials)
            cached_mfa_info = cache.get(f"identity:mfa:{hashed_credentials}")
            otp = self._generate_otp(cached_mfa_info["otp_secret_key"])
            if otp.verify(verify_code):
                return True
            raise ERROR_INVALID_VERIFY_CODE(verify_code=verify_code)

    @staticmethod
    def get_cached_otp_secret_key(credentials: dict):
        if cache.is_set():
            ordered_credentials = OrderedDict(sorted(credentials.items()))
            hashed_credentials = utils.dict_to_hash(ordered_credentials)
            cached_mfa_info = cache.get(f"identity:mfa:{hashed_credentials}")
            cache.delete(f"identity:mfa:{hashed_credentials}")
            return cached_mfa_info["otp_secret_key"]

    @staticmethod
    def _generate_otp_secret_key() -> str:
        return pyotp.random_base32()

    def _generate_otp_qrcode_uri(self, otp, user_id: str) -> str:
        otp_qrcode_uri = otp.provisioning_uri(name=user_id, issuer_name=self.CONST_MFA_OTP_ISSUER_NAME)
        return otp_qrcode_uri

